from pydantic import BaseModel


class GlobalConfig(BaseModel, validate_assignment=True):
    """Global configuration for the generated SDK.

    Args:
        validate_retrieve (bool): Whether to validate the data retrieved from the CDF. Defaults to True.
            Note setting this to False can lead to unexpected behavior if required fields are missing
            or have the wrong type.

    """

    validate_retrieve: bool = True


global_config = GlobalConfig()
